# pylint: disable=wrong-import-position

import sys
import os
import unittest

PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))

from bobot import bobot
from cases.text import simpleText, regexText, arrayText

dev_bot_token = '254968587:AAE1TDdb0f__jl_LDKkpCjtd2eE4UHsTn1Y'

def getBot(rules):
    bot = bobot.init(dev_bot_token)
    bot.setWebhook(None)
    bot.rule(rules)
    return bot

class MyTestCase(unittest.TestCase):
    def testSimpleText(self):
        bot = getBot(simpleText.rules)
        self.assertTrue(simpleText.check(bot))

    def testRegexp(self):
        bot = getBot(regexText.rules)
        self.assertTrue(regexText.check(bot))

    def testArray(self):
        bot = getBot(arrayText.rules)
        self.assertTrue(arrayText.check(bot))

unittest.main()



"""
    User response module
"""

from bobot.utils.utils import flatten
from bobot.Errors import RuleNameError, ResponseFormatError, ResponseMessageError

def createMessage(message, data):
    """
        Creates message from message description object
        and update's data
    """
    options = {}
    text = message

    if isinstance(message, dict):
        text = message.get('text')
        options = message.get('options', {})

        if not text:
            raise ResponseMessageError('Message description should have "text" [String] property', 'text')

        if message.get('interpolate', False):
            text = text.format(
                text=data.get('text'),
                first_name=data.get('first_name'),
                username=data.get('username'),
                second_name=data.get('second_name'))

    return text, options

def processResponseBody(responseBody):
    "Processing response description body to actions lists"

    if isinstance(responseBody, dict):
        return [responseBody]
    elif isinstance(responseBody, list):
        return list(map(processResponseBody, responseBody))
    else:
        raise ResponseFormatError('Wrong response description. Expected dict, got {}'.format(type(responseBody)))

def sendMessage(message=''):
    "Returns message sending function"

    def action(bot, chatId, data):
        "Sends message"
        text, options = createMessage(message, data)
        return bot.send(chatId, text, options)
    return action

def sendMessages(*messages):
    "Returns messages sending function"

    def action(bot, chatId, data):
        "Sends messages"
        for message in messages:
            text, options = createMessage(message, data)
            bot.send(chatId, text, options)
    return action

def sendKeyboard(kb={}):
    "Returns function that sends keyboard to user"
    def action(bot, chatId, data):
        # pylint: disable=unused-argument,missing-docstring
        bot.keyboard(chatId, kb)
    return action

def sendPhoto(photo=None, caption=''):
    "Returns function that sends photo to user"
    if not photo:
        raise ResponseFormatError('Specify photo by URL or File')

    def action(bot, chatId, data):
        # pylint: disable=unused-argument,missing-docstring
        bot.sendPhoto(chatId, photo, caption)

    return action

def sendSticker(stickerId):
    "Returns stickerSender"

    def action(bot, chatId, data):
        # pylint: disable=unused-argument,missing-docstring
        bot.sendSticker(chatId, stickerId)

    return action

class Response():
    "Response class"

    __alowedRules = {
        'sendMessages': sendMessages,
        'sendMessage': sendMessage,
        'sendKeyboard': sendKeyboard,
        'sendSticker': sendSticker,
        'sendPhoto': sendPhoto
    }

    def __addAction(self, actionName, actionArgs):
        if isinstance(actionArgs, list):
            action = self.__alowedRules[actionName](*actionArgs)
        else:
            action = self.__alowedRules[actionName](actionArgs)
        self.actions.append(action)

    def __processResponseDescription(self, response):
        for key in response:
            if key in self.__alowedRules:
                self.__addAction(key, response[key])
            else:
                raise RuleNameError('Response property: "{}" is invalid'.format(key))

    def __init__(self, responseBody):
        self.body = responseBody
        self.actions = []

        responses = flatten(processResponseBody(responseBody))

        for response in responses:
            self.__processResponseDescription(response)

    def run(self, update, bot):
        """
            Runs response actions
        """
        message = update.get('message', {})
        text = message.get('text')
        sender = message.get('from', {})
        senderId = sender.get('id')

        data = {
            'text': text,
            'username': sender.get('username'),
            'first_name': sender.get('first_name'),
            'second_name': sender.get('second_name')
        }

        for action in self.actions:
            action(bot, senderId, data)



"""
	Bobot: Simple Telegram Bot API Wrapper for Python3.5
	https://github.com/zefirka/bobot
"""

from .bobot import init
from .Rule import Rule
from .Parser import Parser
from .Response import Response



"""
    Main bot module
    Includes class Bot
"""

import re
from json import loads, dumps

from bobot.Rule import Rule
from bobot.Response import Response
from bobot.Errors import MessageTextEmptyError
from bobot.utils.req import get, post
from bobot.utils.utils import getFile

__token = None
__api = 'https://api.telegram.org/bot{token}/{method}'

class Bot(object):
    """
        Main Bot class
    """

    __info = None

    def __addRule(self, rule):
        if not isinstance(rule, Rule):
            raise Exception('rule is not Rule instance')

        self.rules.append(rule)
        return self.rules

    #####################################################
    #####################################################

    def __init__(self):
        self.rules = []
        self.clients = {}

    def about(self):
        "Returns information about bot"

        if self.__info:
            return self.__info

        info = call('getMe')
        info = loads(info)
        self.__info = info
        return info.get('result')

    def send(self, chatId, text, options={}):
        """
            Sends text to user
            @public
            @param {str} chatId
            @param  {str} text
            @return {json}
        """

        if not text:
            raise MessageTextEmptyError('Specify message\'s text')

        data = {
            'chat_id': chatId,
            'text': text
        }

        data.update(options)

        return call('sendMessage', data)

    def sendSticker(self, chatId, stickerId):
        """
            Sends photo to user
            @public
            @param {str}        chatId
            @param {str}        strickerId
        """

        data = {
            'chat_id': chatId,
            'sticker': stickerId
        }

        return call('sendSticker', data)

    def sendPhoto(self, chatId, photo, caption=None):
        """
            Sends photo to user
            @public
            @param {str}        chatId
            @param {str|file}   photo
            @param {strp}       [caption]
            @return {tuple}     <photos, result>
        """

        data = {
            'chat_id': chatId,
            'caption': caption
        }

        if isinstance(photo, str):
            photo = getFile(photo)

        file = {'photo': photo}

        res = call('sendPhoto', data, file)
        photos = res.get('result', {}).get('photos', [])

        return photos, res

    def keyboard(self, chatId, text, keyboard=None):
        """
            Sends keyboard to user
            @public
            @param {str} chatId
            @param {str|dict} text - text or keyboard dict
            @param {dict} [keyboard]
        """

        if not keyboard:
            keyboard = text
            text = text.pop('text', '')

        if not text:
            raise MessageTextEmptyError('Specify text message')

        board = dumps(keyboard)

        return call('sendMessage', {
            'chat_id': chatId,
            'text': text,
            'reply_markup': board
        })

    def process(self, update):
        """
            Process update by bot's rules
            @public
            @param {dict} update
        """
        result = []

        if not len(self.rules):
            return None

        for rule in self.rules:
            result.append(rule.execRule(self, update))

        return list(filter(bool, result))

    def rule(self, rules):
        """
            Assign rules to bot
            @public
            @param {list[Rule]|Rule} rules
        """

        if not isinstance(rules, list):
            rules = [rules]

        for rule in rules:
            self.__addRule(rule)

    def on(self, match, response, flags=0):
        "Subscribes to matching"

        if isinstance(response, dict) or isinstance(response, list):
            response = Response(response)

        self.rule(Rule({
            'name': match,
            'match': lambda text: bool(re.compile(match, flags).match(text)),
            'response': response
        }))

    def getUpdates(self, limit=None, offset=None):
        "Call getUpdates method"

        return call('getUpdates', {
            'limit': limit,
            'offset': offset
        })

    def register(self, user, registerInfo={'id': 'id'}):
        "Registers client to bot memory"
        result = {}
        for key in registerInfo:
            result[key] = user.get(registerInfo[key])
        return result

    def setWebhook(self, url, certificate=None):
        """
            Setting up Telegram Webhook for given bot
            @public
            @param {str} url
            @param {str} [certificate]
            @return {json}
        """

        data = {
            'url': url
        }

        if certificate:
            data['certificate'] = certificate

        return call('setWebhook', data)

    def getToken(self):
        "Returns token"

        return __token


def init(token):
    'Initialize'

    global __token
    __token = token

    bot = Bot()
    return bot

def call(method, data={}, files=None):
    "Calls telegram API"
    url = __api.format(token=__token, method=method)

    if files:
        return loads(post(url, data, None, files))

    return get(url, data)



"Module contains test cases for test server"

import re

from bobot.Rule import Rule
from bobot.Parser import Parser

from cases.files import rules as filerules

parser = Parser()

def gt(x):
    return lambda y: y > x

def lt(x):
    return lambda y: y < x

def kb(bot, upd, body):
    # pylint: disable=unused-argument
    "Send keyboard"

    userId = upd.get('message').get('from').get('id')
    bot.keyboard(userId, 'sosi', {
        'resize_keyboard': True,
        'keyboard': [
            [
                {'text': 'alpha'},
                {'text': 'betta'}
            ]
        ],
        'one_time_keyboard': True
    })

testkb = {
    'resize_keyboard': True,
    'text': 'hallo bro',
    'keyboard': [
        [
            {'text': 'alpha'},
            {'text': 'betta'}
        ]
    ],
    'one_time_keyboard': True
}

rules = [
    Rule({
        'match': ['alpha', 'betta'],
        'response': 'alpha or betta was sent'
    }),
    Rule({
        'match': 'action',
        'action': lambda bot, upd, body: bot.send(upd.get('message').get('from').get('id'), 'action completed bro')
    }),
    Rule({
        'name': 'case',
        'match': 'yo-yo',
        'response': [
            {
                'sendMessage': {
                    'text': 'Salam, {username}!',
                    'interpolate': True
                }
            },
            {
                'sendMessages': 'THIS IS SPARTA!!!'.split(' ')
            },
            {
                'sendMessage': 'kase',
                'sendKeyboard': testkb
            }
        ]
    }),
    Rule({
        'match': lambda text: 'parse' in text,
        'parse': parser.parse,
        'response': '{body}'
    }),
    Rule({
        'match': re.compile(r'^regexp'),
        'response': 'regexp included'
    }),
    Rule({
        'parse': lambda text: text.isdigit() and int(text),
        'match': Rule.all(lt(100), gt(50)),
        'response': '{text} less than 100 and greater than 50'
    }),
    Rule({
        'match': Rule.all(re.compile(r'^666'), re.compile(r'.*hell$')),
        'response': 'DAMN HELL!!!'
    }),
    Rule({
        'match': 'keyboard',
        'action': kb
    })
]

def assign(bot):
    bot.on('test', 'responses from on method as string')
    bot.on(r'^test?$', 'responses from on method as regexp')
    bot.on('allahu akbar', {
        'sendMessage': {
            'text': '<pre>И так я \nтоже могу\n:3</pre>',
            'options': {
                'parse_mode': 'html'
            }
        }
    })
    bot.on('jazz', {
        'sendSticker': 'AAQCABNFpoQqAAT0Jx4o7oATy4kvAAIC'
    })
    bot.rule(rules)
    bot.rule(filerules)

    return bot



"Module containes Rule class"

import re

from bobot.Errors import RuleNameError
from bobot.utils.utils import execValue, isFn
from bobot.Response import Response

__retype = type(re.compile('re'))

def getMatcher(match):
    "Return matcher by `match` value of Rule"

    def fn(text):
        "Mathing by function"
        return match(text)

    def lst(text):
        "Mathing by list of mathers"
        for submatch in match:
            submatcher = getMatcher(submatch)

            if submatcher(text):
                return True
        return False

    def eq(text):
        "Mathing by equality"
        return text == match

    def rematch(text):
        "Mathing as regex"
        return bool(match.match(text))

    if isFn(match):
        return fn

    if isinstance(match, list):
        return lst

    if isinstance(match, str) or isinstance(match, int) or isinstance(match, float):
        return eq

    if isinstance(match, __retype):
        return rematch


class Rule(object):
    "Rules class"

    __alowedRules = ['name', 'match', 'response', 'command', 'action', 'parse', 'transform']

    @staticmethod
    def all(*args):
        "Helper method"

        def matcher(text):
            "Matcher"
            for match in args:
                matcher = getMatcher(match)
                if not matcher(text):
                    return False
            return True
        return matcher

    def __init__(self, d):
        self.match = None

        for key in d:
            if key in self.__alowedRules:
                setattr(self, key, d[key])
            else:
                raise RuleNameError('Rule property: "{}" is invalid'.format(key))

        if hasattr(self, 'command'):
            self.addMatching('/' + self.command)

    def execRule(self, bot, update):
        """
            Executes rule

            @public
            @param {Bot} bot
            @param {dict} update
        """
        message = update.get('message', {})
        text = message.get('text')
        sender = message.get('from', {})
        senderId = sender.get('id')
        username = sender.get('username', sender.get('first_name'))

        body = text

        if hasattr(self, 'parse'):
            body = self.parse(body)

        matcher = getMatcher(self.match)

        if matcher(body):
            if hasattr(self, 'transform'):
                body = self.transform(body)

            if hasattr(self, 'register') and not bot.clients.get(senderId):
                bot.clients[senderId] = bot.register(sender, self.register)

            if hasattr(self, 'action'):
                self.action(bot, update, body)

            if hasattr(self, 'response'):
                if isinstance(self.response, Response):
                    return self.response.run(update, bot)
                elif isinstance(self.response, dict) or isinstance(self.response, list):
                    response = Response(self.response)
                    return response.run(update, bot)
                else:
                    response = execValue(self.response, [body, bot])
                    response = response.format(text=text, username=username, body=body)

                    return bot.send(senderId, response)

    def on(self, text, action):
        "Create rule assigned on text"
        if not isinstance(self.match, dict):
            self.match = {
                'value': self.match
            }
            self.match.setdefault('callbacks', {})
            self.match['callbacks'][text] = action

    def addMatching(self, matching):
        "Add new matching"

        if isinstance(self.match, list):
            self.match.append(matching)
        else:
            self.match = [self.match, matching]
